@@ -1,5 +1,5 @@
 // Copyright (C) 2002 Andrew Tridgell
-// Copyright (C) 2009-2019 Joel Rosdahl
+// Copyright (C) 2009-2020 Joel Rosdahl
 //
 // This program is free software; you can redistribute it and/or modify it
 // under the terms of the GNU General Public License as published by the Free
@@ -97,12 +97,11 @@ log_prefix(bool log_updated_time)
 		struct tm tm;
 		struct timeval tv;
 		gettimeofday(&tv, NULL);
-#ifdef __MINGW64_VERSION_MAJOR
-		localtime_r((time_t *)&tv.tv_sec, &tm);
-#else
-		localtime_r(&tv.tv_sec, &tm);
-#endif
-		strftime(timestamp, sizeof(timestamp), "%Y-%m-%dT%H:%M:%S", &tm);
+		if (localtime_r((time_t *)&tv.tv_sec, &tm) != NULL) {
+			strftime(timestamp, sizeof(timestamp), "%Y-%m-%dT%H:%M:%S", &tm);
+		} else {
+			snprintf(timestamp, sizeof(timestamp), "%lu", tv.tv_sec);
+		}
 		snprintf(prefix, sizeof(prefix),
 		         "[%s.%06d %-5d] ", timestamp, (int)tv.tv_usec, (int)getpid());
 	}
@@ -316,18 +315,34 @@ get_umask(void)
 // whether dest will be compressed, and with which compression level. Returns 0
 // on success and -1 on failure. On failure, errno represents the error.
 int
-copy_file(const char *src, const char *dest, int compress_level)
+copy_file(const char *src,
+          const char *dest,
+          int compress_level,
+          bool via_tmp_file)
 {
 	int fd_out;
+	char *tmp_name = NULL;
 	gzFile gz_in = NULL;
 	gzFile gz_out = NULL;
 	int saved_errno = 0;
 
 	// Open destination file.
-	char *tmp_name = x_strdup(dest);
-	fd_out = create_tmp_fd(&tmp_name);
-	cc_log("Copying %s to %s via %s (%scompressed)",
-	       src, dest, tmp_name, compress_level > 0 ? "" : "un");
+	if (via_tmp_file) {
+		tmp_name = x_strdup(dest);
+		fd_out = create_tmp_fd(&tmp_name);
+		cc_log("Copying %s to %s via %s (%scompressed)",
+		       src, dest, tmp_name, compress_level > 0 ? "" : "un");
+	} else {
+		fd_out = open(dest, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0666);
+		if (fd_out == -1) {
+			saved_errno = errno;
+			close(fd_out);
+			errno = saved_errno;
+			return -1;
+		}
+		cc_log("Copying %s to %s (%scompressed)",
+		       src, dest, compress_level > 0 ? "" : "un");
+	}
 
 	// Open source file.
 	int fd_in = open(src, O_RDONLY | O_BINARY);
@@ -411,8 +426,10 @@ copy_file(const char *src, const char *dest, int compress_level)
 			gzclose(gz_out);
 		}
 		close(fd_out);
-		tmp_unlink(tmp_name);
-		free(tmp_name);
+		if (via_tmp_file) {
+			tmp_unlink(tmp_name);
+			free(tmp_name);
+		}
 		return -1;
 	}
 
@@ -434,13 +451,15 @@ copy_file(const char *src, const char *dest, int compress_level)
 		goto error;
 	}
 
-	if (x_rename(tmp_name, dest) == -1) {
-		saved_errno = errno;
-		cc_log("rename error: %s", strerror(saved_errno));
-		goto error;
-	}
+	if (via_tmp_file) {
+		if (x_rename(tmp_name, dest) == -1) {
+			saved_errno = errno;
+			cc_log("rename error: %s", strerror(saved_errno));
+			goto error;
+		}
 
-	free(tmp_name);
+		free(tmp_name);
+	}
 
 	return 0;
 
@@ -454,8 +473,10 @@ error:
 	if (fd_out != -1) {
 		close(fd_out);
 	}
-	tmp_unlink(tmp_name);
-	free(tmp_name);
+	if (via_tmp_file) {
+		tmp_unlink(tmp_name);
+		free(tmp_name);
+	}
 	errno = saved_errno;
 	return -1;
 }
@@ -464,7 +485,7 @@ error:
 int
 move_file(const char *src, const char *dest, int compress_level)
 {
-	int ret = copy_file(src, dest, compress_level);
+	int ret = copy_file(src, dest, compress_level, true);
 	if (ret != -1) {
 		x_unlink(src);
 	}
@@ -704,9 +725,6 @@ format(const char *format, ...)
 	}
 	va_end(ap);
 
-	if (!*ptr) {
-		fatal("Internal error in format");
-	}
 	return ptr;
 }
 
@@ -1065,87 +1083,6 @@ parse_size_with_suffix(const char *str, uint64_t *size)
 	return true;
 }
 
-
-#if !defined(HAVE_REALPATH) && \
-  defined(_WIN32) && \
-  !defined(HAVE_GETFINALPATHNAMEBYHANDLEW)
-static BOOL GetFileNameFromHandle(HANDLE file_handle, TCHAR *filename,
-                                  WORD cch_filename)
-{
-	BOOL success = FALSE;
-
-	// Get the file size.
-	DWORD file_size_hi = 0;
-	DWORD file_size_lo = GetFileSize(file_handle, &file_size_hi);
-	if (file_size_lo == 0 && file_size_hi == 0) {
-		// Cannot map a file with a length of zero.
-		return FALSE;
-	}
-
-	// Create a file mapping object.
-	HANDLE file_map =
-		CreateFileMapping(file_handle, NULL, PAGE_READONLY, 0, 1, NULL);
-	if (!file_map) {
-		return FALSE;
-	}
-
-	// Create a file mapping to get the file name.
-	void *mem = MapViewOfFile(file_map, FILE_MAP_READ, 0, 0, 1);
-	if (mem) {
-		if (GetMappedFileName(GetCurrentProcess(),
-		                      mem,
-		                      filename,
-		                      cch_filename)) {
-			// Translate path with device name to drive letters.
-			TCHAR temp[512];
-			temp[0] = '\0';
-
-			if (GetLogicalDriveStrings(512-1, temp)) {
-				TCHAR name[MAX_PATH];
-				TCHAR drive[3] = TEXT(" :");
-				BOOL found = FALSE;
-				TCHAR *p = temp;
-
-				do {
-					// Copy the drive letter to the template string.
-					*drive = *p;
-
-					// Look up each device name.
-					if (QueryDosDevice(drive, name, MAX_PATH)) {
-						size_t name_len = _tcslen(name);
-						if (name_len < MAX_PATH) {
-							found = _tcsnicmp(filename, name, name_len) == 0
-							        && *(filename + name_len) == _T('\\');
-							if (found) {
-								// Reconstruct filename using temp_file and replace device path
-								// with DOS path.
-								TCHAR temp_file[MAX_PATH];
-								_sntprintf(temp_file,
-								           MAX_PATH - 1,
-								           TEXT("%s%s"),
-								           drive,
-								           filename+name_len);
-								_tcsncpy(filename, temp_file, _tcslen(temp_file));
-							}
-						}
-					}
-
-					// Go to the next NULL character.
-					while (*p++) {
-						// Do nothing.
-					}
-				} while (!found && *p); // End of string.
-			}
-		}
-		success = TRUE;
-		UnmapViewOfFile(mem);
-	}
-
-	CloseHandle(file_map);
-	return success;
-}
-#endif
-
 // A sane realpath() function, trying to cope with stupid path limits and a
 // broken API. Caller frees.
 char *
@@ -1165,12 +1102,13 @@ x_realpath(const char *path)
 		path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
 		FILE_ATTRIBUTE_NORMAL, NULL);
 	if (INVALID_HANDLE_VALUE != path_handle) {
-#ifdef HAVE_GETFINALPATHNAMEBYHANDLEW
-		GetFinalPathNameByHandle(path_handle, ret, maxlen, FILE_NAME_NORMALIZED);
-#else
-		GetFileNameFromHandle(path_handle, ret, maxlen);
-#endif
+		bool ok = GetFinalPathNameByHandle(
+			path_handle, ret, maxlen, FILE_NAME_NORMALIZED) != 0;
 		CloseHandle(path_handle);
+		if (!ok) {
+			free(ret);
+			return x_strdup(path);
+		}
 		p = ret + 4; // Strip \\?\ from the file name.
 	} else {
 		snprintf(ret, maxlen, "%s", path);
@@ -1519,9 +1457,8 @@ x_rename(const char *oldpath, const char *newpath)
 	return rename(oldpath, newpath);
 #else
 	// Windows' rename() refuses to overwrite an existing file.
-	unlink(newpath); // Not x_unlink, as x_unlink calls x_rename.
 	// If the function succeeds, the return value is nonzero.
-	if (MoveFileA(oldpath, newpath) == 0) {
+	if (MoveFileExA(oldpath, newpath, MOVEFILE_REPLACE_EXISTING) == 0) {
 		LPVOID lp_msg_buf;
 		DWORD dw = GetLastError();
 		FormatMessage(
@@ -1575,7 +1512,7 @@ do_x_unlink(const char *path, bool log_failure)
 	// If path is on an NFS share, unlink isn't atomic, so we rename to a temp
 	// file. We don't care if the temp file is trashed, so it's always safe to
 	// unlink it first.
-	char *tmp_name = format("%s.rm.%s", path, tmp_string());
+	char *tmp_name = format("%s.ccache.rm.tmp", path);
 
 	int result = 0;
 	if (x_rename(path, tmp_name) == -1) {
